"""
Authentication can follow one of three paths: the Authorization Code Flow, the Implicit Flow (not currently supported), or the Hybrid Flow (not currently supported).
The flows determine how the ID Token and Access Token are returned to the Relying Party (client).

For details see `Section 3 of OpenID Connect Core 1.0 <https://openid.net/specs/openid-connect-core-1_0.html#Authentication>`_
"""
